import sys
from collections.abc import Callable
from dataclasses import dataclass
from functools import partial
from inspect import signature
from typing import ClassVar
from unittest.mock import Mock

import pytest

from psygnal import EmissionInfo, Signal, SignalGroupDescriptor, SignalInstance, evented
from psygnal._group import PathStep

if all(x not in {"--codspeed", "--benchmark", "tests/test_bench.py"} for x in sys.argv):
    pytest.skip("use --benchmark to run benchmark", allow_module_level=True)

CALLBACK_TYPES = [
    "function",
    "method",
    "lambda",
    "partial",
    "partial_method",
    "setattr",
    "setitem",
    "real_func",
    "print",
]


# fmt: off
class Emitter:
    one_int = Signal(int)
    int_str = Signal(int, str)


class Obj:
    x: int = 0
    def __setitem__(self, key: str, value: int) -> None:
        self.x = value
    def no_args(self) -> None: ...
    def one_int(self, x: int) -> None: ...
    def int_str(self, x: int, y: str) -> None: ...

def no_args() -> None: ...
def one_int(x: int) -> None: ...
def int_str(x: int, y: str) -> None: ...
def real_func() -> None: list(range(4))  # simulate a brief thing

INT_SIG = signature(one_int)
# fmt: on


def _get_callback(callback_type: str, obj: Obj) -> Callable:
    callback_types: dict[str, Callable] = {
        "function": one_int,
        "method": obj.one_int,
        "lambda": lambda x: None,
        "partial": partial(int_str, y="foo"),
        "partial_method": partial(obj.int_str, y="foo"),
        "real_func": real_func,
        "print": print,
    }
    return callback_types[callback_type]


# Creation suite ------------------------------------------


def test_create_signal(benchmark: Callable) -> None:
    benchmark(Signal, int)


def test_create_signal_instance(benchmark: Callable) -> None:
    benchmark(SignalInstance, INT_SIG)


# Connect suite ---------------------------------------------


@pytest.mark.parametrize("check_types", ["check_types", ""])
@pytest.mark.parametrize("callback_type", CALLBACK_TYPES)
def test_connect_time(
    benchmark: Callable, callback_type: str, check_types: str
) -> None:
    emitter = Emitter()
    obj = Obj()
    kwargs = {}
    if callback_type == "setattr":
        func: Callable = emitter.one_int.connect_setattr
        args: tuple = (obj, "x")
        kwargs = {"maxargs": 1}
    elif callback_type == "setitem":
        func = emitter.one_int.connect_setitem
        args = (obj, "x")
        kwargs = {"maxargs": 1}
    else:
        func = emitter.one_int.connect
        args = (_get_callback(callback_type, obj),)
        kwargs = {"check_types": bool(check_types)}

    benchmark(func, *args, **kwargs)


# Emit suite ------------------------------------------------


@pytest.mark.parametrize("n_connections", range(2, 2**6, 16))
@pytest.mark.parametrize("callback_type", CALLBACK_TYPES)
def test_emit_time(benchmark: Callable, n_connections: int, callback_type: str) -> None:
    emitter = Emitter()
    obj = Obj()
    if callback_type == "setattr":
        for _ in range(n_connections):
            emitter.one_int.connect_setattr(obj, "x", maxargs=1)
    elif callback_type == "setitem":
        for _ in range(n_connections):
            emitter.one_int.connect_setitem(obj, "x", maxargs=1)
    else:
        callback = _get_callback(callback_type, obj)
        for _ in range(n_connections):
            emitter.one_int.connect(callback, unique=False)

    benchmark(emitter.one_int.emit, 1)


def test_emit_fast(benchmark: Callable) -> None:
    emitter = Emitter()
    emitter.one_int.connect(one_int)
    benchmark(emitter.one_int.emit_fast, 1)


@pytest.mark.benchmark
def test_evented_creation() -> None:
    @evented
    @dataclass
    class Obj:
        x: int = 0
        y: str = "hi"
        z: bool = False

    _ = Obj().events  # type: ignore


def test_evented_setattr(benchmark: Callable) -> None:
    @evented
    @dataclass
    class Obj:
        x: int = 0
        y: str = "hi"
        z: bool = False

    obj = Obj()
    _ = obj.events  # type: ignore

    benchmark(setattr, obj, "x", 1)


def _get_dataclass(type_: str) -> type:
    if type_ == "attrs":
        from attrs import define

        @define
        class Foo:
            a: int
            b: str
            c: bool
            d: float
            e: tuple[int, str]
            events: ClassVar = SignalGroupDescriptor()

    elif type_ == "dataclass":

        @dataclass
        class Foo:  # type: ignore [no-redef]
            a: int
            b: str
            c: bool
            d: float
            e: tuple[int, str]
            events: ClassVar = SignalGroupDescriptor()

    elif type_ == "msgspec":
        import msgspec

        class Foo(msgspec.Struct):  # type: ignore [no-redef]
            a: int
            b: str
            c: bool
            d: float
            e: tuple[int, str]
            events: ClassVar = SignalGroupDescriptor()

    elif type_ == "pydantic":
        from pydantic import BaseModel

        class Foo(BaseModel):  # type: ignore [no-redef]
            a: int
            b: str
            c: bool
            d: float
            e: tuple[int, str]
            events: ClassVar = SignalGroupDescriptor()

    return Foo


@pytest.mark.parametrize("type_", ["dataclass", "pydantic", "attrs", "msgspec"])
def test_dataclass_group_create(type_: str, benchmark: Callable) -> None:
    if type_ == "msgspec":
        pytest.importorskip("msgspec")

    Foo = _get_dataclass(type_)
    foo = Foo(a=1, b="hi", c=True, d=1.0, e=(1, "hi"))
    benchmark(getattr, foo, "events")


@pytest.mark.parametrize("type_", ["dataclass", "pydantic", "attrs", "msgspec"])
def test_dataclass_setattr(type_: str, benchmark: Callable) -> None:
    if type_ == "msgspec":
        pytest.importorskip("msgspec")

    Foo = _get_dataclass(type_)
    foo = Foo(a=1, b="hi", c=True, d=1.0, e=(1, "hi"))
    mock = Mock()
    foo.events._psygnal_relay.connect(mock)

    def _doit() -> None:
        foo.a = 2
        foo.b = "hello"
        foo.c = False
        foo.d = 2.0
        foo.e = (2, "hello")

    benchmark(_doit)
    for emitted, attr in zip(
        [(2, 1), ("hello", "hi"), (False, True), (2.0, 1.0), ((2, "hello"), (1, "hi"))],
        "abcde",
        strict=False,
    ):
        mock.assert_any_call(
            EmissionInfo(getattr(foo.events, attr), emitted, (PathStep(attr=attr),))
        )
        assert getattr(foo, attr) == emitted[0]



from dataclasses import dataclass, field
from typing import ClassVar, cast
from unittest.mock import Mock

import pytest

from psygnal import EmissionInfo, PathStep, SignalGroupDescriptor, evented
from psygnal.containers import EventedDict, EventedList, EventedSet

EL2 = EventedList([10, 20, 30])
EL = EventedList([1, 2, EL2])  # TODO: re-emit events from nested lists
ED: EventedDict[str, int] = EventedDict({"a": 1, "b": 2})
ES = EventedSet({"x", "y", "z"})


@evented
@dataclass
class A:
    x: int = 1


@pytest.fixture
def EventedClass():
    @dataclass
    class M:
        events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor(
            connect_child_events=True
        )
        a: EventedList = field(default_factory=lambda: EL.copy())
        b: EventedDict = field(default_factory=lambda: ED.copy())
        c: EventedSet = field(default_factory=lambda: ES.copy())
        d: A = field(default_factory=A)

    return M


@pytest.mark.parametrize(
    "expr, expect",
    [
        # list element reassignment
        ("m.a[1] = 12", ((1, 2, 12), (PathStep(attr="a"), PathStep(index=1)))),
        # list attribute replaced wholesale
        ("m.a = [0, 2]", (([0, 2], EL), (PathStep(attr="a"),))),
        # dict item updated
        ("m.b['a'] = 3", (("a", 1, 3), (PathStep(attr="b"), PathStep(key="a")))),
        # dataclass attribute replaced
        ("m.b = {'x': 11}", (({"x": 11}, ED), (PathStep(attr="b"),))),
        # set mutated
        ("m.c.add('w')", ((("w",), ()), (PathStep(attr="c"),))),
        # set attribute replaced wholesale
        (r"m.c = {1}", (({1}, ES), (PathStep(attr="c"),))),
        # nested dataclass field change
        ("m.d.x = 12", ((12, 1), (PathStep(attr="d"), PathStep(attr="x")))),
        # dataclass attribute replaced wholesale
        ("m.d = {'x': 1}", (({"x": 1}, A(x=1)), (PathStep(attr="d"),))),
        # evented object replaced with another evented object
        ("m.d = A(x=99)", ((A(x=99), A(x=1)), (PathStep(attr="d"),))),
    ],
)
def test_nested_containers(expr, expect, EventedClass):
    m = EventedClass()

    mock = Mock()
    m.events.connect(mock)

    exec(expr)
    info = cast("EmissionInfo", mock.call_args[0][0])
    assert (info.args, info.path) == expect



from contextlib import nullcontext
from dataclasses import dataclass
from typing import Any, ClassVar
from unittest.mock import Mock, patch

import pytest

from psygnal import (
    Signal,
    SignalGroup,
    SignalGroupDescriptor,
    _compiled,
    _group_descriptor,
)


class MyGroup(SignalGroup):
    sig = Signal()


@pytest.mark.parametrize("type_", ["dataclass", "pydantic", "attrs", "msgspec"])
def test_descriptor_inherits(type_: str) -> None:
    if type_ == "dataclass":
        from dataclasses import dataclass

        @dataclass
        class Base:
            a: int
            events: ClassVar = SignalGroupDescriptor()

        @dataclass
        class Foo(Base):
            b: str

        @dataclass
        class Bar(Foo):
            c: float

    elif type_ == "pydantic":
        pytest.importorskip("pydantic")
        from pydantic import BaseModel

        class Base(BaseModel):
            a: int
            events: ClassVar = SignalGroupDescriptor()

        class Foo(Base):
            b: str

        class Bar(Foo):
            c: float

    elif type_ == "attrs":
        from attrs import define

        @define
        class Base:
            a: int
            events: ClassVar = SignalGroupDescriptor()

        @define
        class Foo(Base):
            b: str

        @define
        class Bar(Foo):
            c: float

    elif type_ == "msgspec":
        msgspec = pytest.importorskip("msgspec")

        class Base(msgspec.Struct):  # type: ignore
            a: int
            events: ClassVar = SignalGroupDescriptor()

        class Foo(Base):
            b: str

        class Bar(Foo):
            c: float

    assert Bar.events is Base.events

    with patch.object(
        _group_descriptor, "evented_setattr", wraps=_group_descriptor.evented_setattr
    ) as mock_decorator:
        base = Base(a=1)
        foo = Foo(a=1, b="2")
        bar = Bar(a=1, b="2", c=3.0)
        bar2 = Bar(a=1, b="2", c=3.0)

        # the patching of __setattr__ should only happen once
        # and it will happen only on the first access of .events
        mock_decorator.assert_not_called()
        assert set(base.events) == {"a"}
        assert set(foo.events) == {"a", "b"}
        assert set(bar.events) == {"a", "b", "c"}
        assert set(bar2.events) == {"a", "b", "c"}
        if not _compiled:  # can't patch otherwise
            assert mock_decorator.call_count == 1

    mock = Mock()
    foo.events.a.connect(mock)

    # base doesn't affect subclass
    base.events.a.emit(1)
    mock.assert_not_called()

    # subclass doesn't affect superclass
    bar.events.a.emit(1)
    mock.assert_not_called()

    foo.events.a.emit(1)
    mock.assert_called_once_with(1)


@pytest.mark.parametrize("patch_setattr", [True, False])
def test_no_patching(patch_setattr: bool) -> None:
    """Test patch_setattr=False doesn't patch the class"""

    # sourcery skip: extract-duplicate-method
    @dataclass
    class Foo:
        a: int
        _events: ClassVar = SignalGroupDescriptor(patch_setattr=patch_setattr)

    with patch.object(
        _group_descriptor, "evented_setattr", wraps=_group_descriptor.evented_setattr
    ) as mock_decorator:
        foo = Foo(a=1)
        _ = foo._events
        if not _compiled:  # can't patch otherwise
            assert mock_decorator.call_count == int(patch_setattr)

    assert _group_descriptor.is_evented(Foo.__setattr__) == patch_setattr
    mock = Mock()
    foo._events.a.connect(mock)
    foo.a = 2
    if patch_setattr:
        mock.assert_called_once_with(2, 1)
    else:
        mock.assert_not_called()


def test_direct_patching() -> None:
    """Test directly using evented_setattr on a class"""
    mock1 = Mock()

    @dataclass
    class Foo:
        a: int
        _events: ClassVar = SignalGroupDescriptor(patch_setattr=False)

        @_group_descriptor.evented_setattr("_events")
        def __setattr__(self, __name: str, __value: Any) -> None:
            old = getattr(self, __name, None)
            mock1(__name, __value, old)
            super().__setattr__(__name, __value)

    assert _group_descriptor.is_evented(Foo.__setattr__)

    # patch again ... this should NOT cause a double event emission.
    Foo.__setattr__ = _group_descriptor.evented_setattr("_events", Foo.__setattr__)

    foo = Foo(a=1)
    mock = Mock()
    foo._events.a.connect(mock)
    foo.a = 2
    mock.assert_called_once_with(2, 1)  # confirm no double event emission
    mock1.assert_called_with("a", 2, 1)


def test_no_getattr_on_non_evented_fields() -> None:
    """Make sure that we're not accidentally calling getattr on non-evented fields."""
    a_mock = Mock()
    b_mock = Mock()

    @dataclass
    class Foo:
        a: int
        events: ClassVar = SignalGroupDescriptor()

        @property
        def b(self) -> int:
            b_mock(self._b)
            return self._b

        @b.setter
        def b(self, value: int) -> None:
            self._b = value

    foo = Foo(a=1)
    foo.events.a.connect(a_mock)
    foo.a = 2
    a_mock.assert_called_once_with(2, 1)

    foo.b = 1
    b_mock.assert_not_called()  # getter shouldn't have been called
    assert foo.b == 1
    b_mock.assert_called_once_with(1)  # getter should have been called only once


def test_evented_field_connect_setattr() -> None:
    """Test that using connect_setattr"""

    @dataclass
    class Foo:
        a: int
        events: ClassVar = SignalGroupDescriptor()

    class Bar:
        x = 1
        y = 1

    foo = Foo(a=1)
    bar = Bar()

    foo.events.a.connect_setattr(bar, "x")
    foo.events.a.connect_setattr(bar, "y", maxargs=None)
    foo.events.a.emit(2, 1)

    assert bar.x == 2  # this is likely the desired outcome
    # this is a bit of a gotcha, but it's the expected behavior
    # when using connect_setattr with maxargs=None
    # remove this test if/when we change maxargs to default to 1 on SignalInstance
    assert bar.y == (2, 1)  # type: ignore


@pytest.mark.parametrize("collect", [True, False])
@pytest.mark.parametrize("klass", [None, SignalGroup, MyGroup])
def test_collect_fields(collect: bool, klass: type[SignalGroup] | None) -> None:
    signal_class = klass or SignalGroup
    should_fail_def = signal_class is SignalGroup and collect is False
    ctx = pytest.raises(ValueError) if should_fail_def else nullcontext()

    with ctx:

        @dataclass
        class Foo:
            events: ClassVar = SignalGroupDescriptor(
                warn_on_no_fields=False,
                signal_group_class=klass,
                collect_fields=collect,
            )
            a: int = 1

    if should_fail_def:
        return

    @dataclass
    class Bar(Foo):
        b: float = 2.0

    foo = Foo()
    bar = Bar()

    assert issubclass(type(foo.events), signal_class)

    if collect:
        assert type(foo.events) is not signal_class
        assert "a" in foo.events
        assert "a" in bar.events
        assert "b" in bar.events

    else:
        assert type(foo.events) is signal_class
        assert "a" not in foo.events
        assert "a" not in bar.events
        assert "b" not in bar.events

    if signal_class is MyGroup:
        assert "sig" in foo.events
        assert "sig" in bar.events



from __future__ import annotations

import asyncio
import gc
import importlib.util
import signal
from typing import TYPE_CHECKING, Any, Literal, Protocol
from unittest.mock import Mock

import pytest
import pytest_asyncio

from psygnal import _async
from psygnal._weak_callback import WeakCallback, weak_callback

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

# Available backends for parametrization
AVAILABLE_BACKENDS = ["asyncio"]
if importlib.util.find_spec("trio") is not None:
    AVAILABLE_BACKENDS.append("trio")
if importlib.util.find_spec("anyio") is not None:
    AVAILABLE_BACKENDS.append("anyio")


class BackendTestRunner(Protocol):
    """Protocol for backend-specific test runners."""

    @property
    def backend_name(self) -> Literal["asyncio", "anyio", "trio"]:
        """Name of the backend being used."""
        ...

    async def sleep(self, duration: float) -> None:
        """Sleep for the given duration using backend-specific sleep."""
        ...

    def run_with_backend(self, test_func: Callable[[], Any]) -> Any:
        """Run a test function with proper backend setup and teardown. Synchronous."""
        ...


class AsyncioTestRunner:
    """Test runner for asyncio backend."""

    @property
    def backend_name(self) -> Literal["asyncio"]:
        return "asyncio"

    async def sleep(self, duration: float) -> None:
        await asyncio.sleep(duration)

    def run_with_backend(self, test_func: Callable[[], Any]) -> Any:
        """Run test with asyncio backend."""

        async def _run_test() -> Any:
            _async.clear_async_backend()
            backend = _async.set_async_backend("asyncio")

            # Wait for backend to be running
            await self._wait_for_backend_running(backend)

            try:
                return await test_func()
            finally:
                # Cleanup
                if hasattr(backend, "_task") and not backend._task.done():
                    backend._task.cancel()
                    try:
                        await backend._task
                    except asyncio.CancelledError:
                        pass
                _async.clear_async_backend()

        return asyncio.run(_run_test())

    async def _wait_for_backend_running(
        self, backend: _async._AsyncBackend, timeout: float = 1.0
    ) -> None:
        """Wait for backend to be running with a timeout."""
        start_time = asyncio.get_event_loop().time()
        while not backend.running.is_set():
            if asyncio.get_event_loop().time() - start_time > timeout:
                raise TimeoutError("Backend did not start running within timeout")
            await asyncio.sleep(0)


class AnyioTestRunner:
    """Test runner for anyio backend."""

    @property
    def backend_name(self) -> Literal["anyio"]:
        return "anyio"

    async def sleep(self, duration: float) -> None:
        import anyio

        await anyio.sleep(duration)

    def run_with_backend(self, test_func: Callable[[], Any]) -> Any:
        """Run test with anyio backend using structured concurrency."""
        import anyio

        async def _run_test():
            _async.clear_async_backend()
            backend = _async.set_async_backend("anyio")

            result = None
            async with anyio.create_task_group() as tg:
                tg.start_soon(backend.run)

                # Wait for backend to be running
                await backend.running.wait()

                try:
                    result = await test_func()
                finally:
                    # Cancel task group to shutdown properly
                    tg.cancel_scope.cancel()

            _async.clear_async_backend()
            return result

        return anyio.run(_run_test)


class TrioTestRunner:
    """Test runner for trio backend."""

    @property
    def backend_name(self) -> Literal["trio"]:
        return "trio"

    async def sleep(self, duration: float) -> None:
        import trio

        await trio.sleep(duration)

    def run_with_backend(self, test_func: Callable[[], Any]) -> Any:
        """Run test with trio backend using structured concurrency."""

        # On Windows asyncio has probably left its FD installed
        try:
            signal.set_wakeup_fd(-1)  # restore default
        except (ValueError, AttributeError):
            pass  # not the main thread or not supported

        import trio

        result = None

        async def _trio_main():
            nonlocal result
            _async.clear_async_backend()
            backend = _async.set_async_backend("trio")

            # Use a timeout to prevent hanging
            with trio.move_on_after(5.0) as cancel_scope:
                async with trio.open_nursery() as nursery:
                    nursery.start_soon(backend.run)

                    # Wait for backend to be running
                    await backend.running.wait()

                    try:
                        result = await test_func()
                    finally:
                        # Cancel nursery to shutdown properly
                        nursery.cancel_scope.cancel()

            # Check if we timed out
            if cancel_scope.cancelled_caught:
                raise TimeoutError("Test timed out")

            _async.clear_async_backend()

        # Run in trio context
        trio.run(_trio_main)
        return result


async def mock_call_count(
    mock: Mock, runner: BackendTestRunner, max_iterations: int = 100
) -> None:
    """Wait for callback execution with backend-specific sleep."""
    for _ in range(max_iterations):
        await runner.sleep(0.01)
        if mock.call_count > 0:
            break


@pytest_asyncio.fixture
async def clean_async_backend():
    """Fixture to ensure clean async backend state."""
    _async.clear_async_backend()
    yield
    _async.clear_async_backend()


@pytest.fixture(params=AVAILABLE_BACKENDS)
def runner(
    request: pytest.FixtureRequest, clean_async_backend: None
) -> Iterator[BackendTestRunner]:
    """Get the backend runner for the specified backend."""
    mapping: dict[str, type[BackendTestRunner]] = {
        "asyncio": AsyncioTestRunner,
        "anyio": AnyioTestRunner,
        "trio": TrioTestRunner,
    }
    yield mapping[request.param]()


# Parametrized tests for all backends
@pytest.mark.parametrize(
    "slot_type",
    [
        "coroutinefunc",
        "weak_coroutinefunc",
        "coroutinemethod",
    ],
)
def test_slot_types_all_backends(runner: BackendTestRunner, slot_type: str) -> None:
    """Test async slot types with all available backends."""

    async def _test_slot_type():
        mock = Mock()
        final_mock = Mock()

        if slot_type in {"coroutinefunc", "weak_coroutinefunc"}:

            async def test_obj(x: int) -> int:
                mock(x)
                return x

            cb = weak_callback(
                test_obj,
                strong_func=(slot_type == "coroutinefunc"),
                finalize=final_mock,
            )
        elif slot_type == "coroutinemethod":

            class MyObj:
                async def coroutine_method(self, x: int) -> int:
                    mock(x)
                    return x

            obj = MyObj()
            cb = weak_callback(obj.coroutine_method, finalize=final_mock)

        assert isinstance(cb, WeakCallback)
        assert isinstance(cb.slot_repr(), str)
        assert cb.dereference() is not None

        # Test callback execution
        cb.cb((2,))
        await mock_call_count(mock, runner)
        mock.assert_called_once_with(2)

        # Test direct await
        mock.reset_mock()
        result = await cb(4)
        assert result == 4
        mock.assert_called_once_with(4)

        # Test weak reference cleanup
        if slot_type in {"coroutinefunc", "weak_coroutinefunc"}:
            del test_obj
        else:
            del obj
        gc.collect()

        if slot_type == "coroutinefunc":  # strong_func
            cb.cb((4,))
            await mock_call_count(mock, runner)
            mock.assert_called_with(4)
        else:
            await mock_call_count(final_mock, runner)
            final_mock.assert_called_once_with(cb)
            assert cb.dereference() is None
            with pytest.raises(ReferenceError):
                cb.cb((2,))
            with pytest.raises(ReferenceError):
                await cb(2)

    # Run the test with the appropriate backend
    runner.run_with_backend(_test_slot_type)


def test_backend_error_conditions(runner: BackendTestRunner) -> None:
    """Test backend error conditions and exception handling."""

    async def _test_errors():
        mock = Mock()

        async def test_coro(x: int) -> int:
            if x == 999:
                raise ValueError("Test error")
            mock(x)
            return x

        cb = weak_callback(test_coro, strong_func=True)

        # Test normal execution
        cb.cb((5,))
        await mock_call_count(mock, runner)
        mock.assert_called_once_with(5)

        # Test error case - should not crash the backend
        cb.cb((999,))
        await runner.sleep(0.1)  # Give time for error to be handled

        # Backend should still work after error
        mock.reset_mock()
        cb.cb((10,))
        await mock_call_count(mock, runner)
        mock.assert_called_once_with(10)

    # Run the test with the backend runner
    runner.run_with_backend(_test_errors)


@pytest.mark.usefixtures("clean_async_backend")
@pytest.mark.asyncio
async def test_run_method_early_return() -> None:
    """Test that run() method returns early if backend is already running."""
    backend = _async.set_async_backend("asyncio")

    # Wait for backend to be running
    start_time = asyncio.get_event_loop().time()
    while not backend.running.is_set():
        if asyncio.get_event_loop().time() - start_time > 1.0:
            raise TimeoutError("Backend did not start running within timeout")
        await asyncio.sleep(0)

    # Now calling run() again should return early
    await backend.run()

    # Backend should still be running
    assert backend.running.is_set()


@pytest.mark.parametrize("backend_name", AVAILABLE_BACKENDS)
def test_high_level_api(backend_name: Literal["trio", "asyncio", "anyio"]) -> None:
    """Test the exact usage pattern shown in the feature summary documentation."""

    def run_example() -> None:
        """The example from the feature summary, adapted for testing."""

        async def example_main() -> None:
            # Step 1: Set Backend Early (Once Per Application)
            backend = _async.set_async_backend(backend_name)

            # Step 2: Launch Backend in Your Event Loop (backend-specific)
            if backend_name == "asyncio":
                import asyncio

                # Start the backend as a background task
                async_backend = _async.get_async_backend()
                assert async_backend is not None
                task = asyncio.create_task(async_backend.run())

                # Wait for backend to be running
                await backend.running.wait()

            elif backend_name == "anyio":
                import anyio

                async with anyio.create_task_group() as tg:
                    # Start the backend in the task group
                    async_backend = _async.get_async_backend()
                    assert async_backend is not None
                    tg.start_soon(async_backend.run)

                    # Wait for backend to be running
                    await backend.running.wait()

                    # Run the actual example
                    await run_signal_example()

                    # Cancel to exit cleanly
                    tg.cancel_scope.cancel()
                return

            elif backend_name == "trio":
                import trio

                async with trio.open_nursery() as nursery:
                    # Start the backend in the nursery
                    async_backend = _async.get_async_backend()
                    assert async_backend is not None
                    nursery.start_soon(async_backend.run)

                    # Wait for backend to be running
                    await backend.running.wait()

                    # Run the actual example
                    await run_signal_example()

                    # Cancel to exit cleanly
                    nursery.cancel_scope.cancel()
                return

            # For asyncio, run the example after backend is started
            try:
                await run_signal_example()
            finally:
                if backend_name == "asyncio":
                    task.cancel()
                    try:
                        await task
                    except asyncio.CancelledError:
                        pass

        async def run_signal_example() -> None:
            """Step 3: Connect Async Callbacks - the exact example from docs."""
            from psygnal import Signal

            class MyObj:
                value_changed = Signal(str)

                def set_value(self, value: str) -> None:
                    self.value_changed.emit(value)

            # Track calls for testing
            mock = Mock()

            async def on_value_changed(new_value: str) -> None:
                mock(new_value)

            obj = MyObj()
            obj.value_changed.connect(on_value_changed)
            obj.set_value("hello!")

            # Wait for callback to execute
            max_wait = 100
            for _ in range(max_wait):
                if mock.call_count > 0:
                    break
                if backend_name == "asyncio":
                    await asyncio.sleep(0.01)
                elif backend_name == "anyio":
                    import anyio

                    await anyio.sleep(0.01)
                elif backend_name == "trio":
                    import trio

                    await trio.sleep(0.01)

            # Verify the callback was called with the expected value
            assert mock.call_count == 1
            assert mock.call_args[0][0] == "hello!"

        # Run the example with the appropriate backend
        if backend_name == "asyncio":
            return asyncio.run(example_main())
        elif backend_name == "anyio":
            import anyio

            return anyio.run(example_main)
        elif backend_name == "trio":
            import trio

            return trio.run(example_main)

    # Clear any existing backend before test
    _async.clear_async_backend()
    try:
        run_example()
    finally:
        # Clean up after test
        _async.clear_async_backend()



from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Annotated
from unittest.mock import Mock, call

import pytest

import psygnal
from psygnal import EmissionInfo, Signal, SignalGroup, SignalInstance
from psygnal._group import SignalRelay

if TYPE_CHECKING:
    from collections.abc import Callable


class MyGroup(SignalGroup):
    sig1 = Signal(int)
    sig2 = Signal(str)


with pytest.warns():

    class ConflictGroup(SignalGroup):
        sig1 = Signal(int)
        connect = Signal(int)  # type: ignore


def test_cannot_instantiate_group() -> None:
    with pytest.raises(TypeError, match="Cannot instantiate `SignalGroup` directly"):
        SignalGroup()


def test_signal_group() -> None:
    assert not MyGroup.psygnals_uniform()
    with pytest.warns(
        FutureWarning, match="The `is_uniform` method on SignalGroup is deprecated"
    ):
        assert not MyGroup.is_uniform()
    group = MyGroup()
    assert not group.psygnals_uniform()
    assert list(group) == ["sig1", "sig2"]  # testing __iter__
    assert group.sig1 is group["sig1"]
    assert set(group.signals) == {"sig1", "sig2"}

    assert repr(group) == "<SignalGroup 'MyGroup' with 2 signals>"


def test_uniform_group() -> None:
    """In a uniform group, all signals must have the same signature."""

    class MyStrictGroup(SignalGroup, strict=True):
        sig1 = Signal(int)
        sig2 = Signal(int)

    assert MyStrictGroup.psygnals_uniform()
    group = MyStrictGroup()
    assert group.psygnals_uniform()
    assert set(group) == {"sig1", "sig2"}

    with pytest.raises(TypeError) as e:

        class BadGroup(SignalGroup, strict=True):
            sig1 = Signal(str)
            sig2 = Signal(int)

    assert str(e.value).startswith("All Signals in a strict SignalGroup must")


@pytest.mark.skipif(Annotated is None, reason="requires typing.Annotated")
def test_nonhashable_args() -> None:
    """Test that non-hashable annotations are allowed in a SignalGroup"""

    class MyGroup(SignalGroup):
        sig1 = Signal(Annotated[int, {"a": 1}])  # type: ignore
        sig2 = Signal(Annotated[float, {"b": 1}])  # type: ignore

    assert not MyGroup.psygnals_uniform()

    with pytest.raises(TypeError):

        class MyGroup2(SignalGroup, strict=True):
            sig1 = Signal(Annotated[int, {"a": 1}])  # type: ignore
            sig2 = Signal(Annotated[float, {"b": 1}])  # type: ignore


@pytest.mark.parametrize("direct", [True, False])
def test_signal_group_connect(direct: bool) -> None:
    mock = Mock()
    group = MyGroup()
    if direct:
        # the callback wants the emitted arguments directly
        group.connect_direct(mock)
    else:
        # the callback will receive an EmissionInfo tuple
        # (SignalInstance, arg_tuple)
        group.connect(mock)
    group.sig1.emit(1)
    group.sig2.emit("hi")

    assert mock.call_count == 2
    # if connect_with_info was used, the callback will be given an EmissionInfo
    # tuple that contains the args as well as the signal instance used
    if direct:
        expected_calls = [call(1), call("hi")]
    else:
        expected_calls = [
            call(EmissionInfo(group.sig1, (1,))),
            call(EmissionInfo(group.sig2, ("hi",))),
        ]
    mock.assert_has_calls(expected_calls)


def test_signal_group_connect_no_args() -> None:
    """Test that group.all.connect can take a callback that wants no args"""
    group = MyGroup()
    count = []

    def my_slot() -> None:
        count.append(1)

    group.connect(my_slot)
    group.sig1.emit(1)
    group.sig2.emit("hi")
    assert len(count) == 2


def test_group_blocked() -> None:
    group = MyGroup()

    mock1 = Mock()
    mock2 = Mock()

    group.connect(mock1)
    group.sig1.connect(mock2)
    group.sig1.emit(1)

    mock1.assert_called_once_with(EmissionInfo(group.sig1, (1,)))
    mock2.assert_called_once_with(1)

    mock1.reset_mock()
    mock2.reset_mock()

    group.sig2.block()
    assert group.sig2._is_blocked

    with group.all.blocked():
        group.sig1.emit(1)
        assert group.sig1._is_blocked

    assert not group.sig1._is_blocked
    # the blocker should have restored subthings to their previous states
    assert group.sig2._is_blocked

    mock1.assert_not_called()
    mock2.assert_not_called()


def test_group_blocked_exclude() -> None:
    """Test that we can exempt certain signals from being blocked."""
    group = MyGroup()

    mock1 = Mock()
    mock2 = Mock()

    group.sig1.connect(mock1)
    group.sig2.connect(mock2)

    with group.all.blocked(exclude=("sig2",)):
        group.sig1.emit(1)
        group.sig2.emit("hi")
    mock1.assert_not_called()
    mock2.assert_called_once_with("hi")


def test_group_disconnect_single_slot() -> None:
    """Test that we can disconnect single slots from groups."""
    group = MyGroup()

    mock1 = Mock()
    mock2 = Mock()

    group.sig1.connect(mock1)
    group.sig2.connect(mock2)

    group.disconnect(mock1)
    group.sig1.emit()
    mock1.assert_not_called()

    group.sig2.emit()
    mock2.assert_called_once()


def test_group_disconnect_all_slots() -> None:
    """Test that we can disconnect all slots from groups."""
    group = MyGroup()

    mock1 = Mock()
    mock2 = Mock()

    group.sig1.connect(mock1)
    group.sig2.connect(mock2)

    group.disconnect()
    group.sig1.emit()
    group.sig2.emit()

    mock1.assert_not_called()
    mock2.assert_not_called()


def test_weakref() -> None:
    """Make sure that the group doesn't keep a strong reference to the instance."""
    import gc

    class T: ...

    obj = T()
    group = MyGroup(obj)
    assert group.all.instance is obj
    del obj
    gc.collect()
    assert group.all.instance is None


@pytest.mark.parametrize(
    "Group, signame, get_sig",
    [
        (MyGroup, "sig1", getattr),
        (MyGroup, "sig1", SignalGroup.__getitem__),
        (ConflictGroup, "sig1", getattr),
        (ConflictGroup, "sig1", SignalGroup.__getitem__),
        (ConflictGroup, "connect", SignalGroup.__getitem__),
    ],
)
def test_group_deepcopy(
    Group: type[SignalGroup], signame: str, get_sig: Callable
) -> None:
    """_summary_

    Parameters
    ----------
    Group : type[SignalGroup]
        The group class to test, where ConflictGroup has a signal named "connect"
        which conflicts with the SignalGroup method of the same name.
    signame : str
        The name of the signal to test
    get_sig : Callable
        The method to use to get the signal instance from the group. we don't
        test getattr for ConflictGroup because it has a signal named "connect"
    """

    class T:
        def method(self) -> None: ...

    obj = T()
    group = Group(obj)
    assert deepcopy(group) is not group  # but no warning

    group.connect(obj.method)
    group2 = deepcopy(group)

    assert not len(group2.all)
    mock = Mock()
    mock2 = Mock()
    group.connect(mock)
    group2.connect(mock2)

    # test that we can access signalinstances (either using getattr or __getitem__)
    siginst1 = get_sig(group, signame)
    siginst2 = get_sig(group2, signame)
    assert isinstance(siginst1, SignalInstance)
    assert isinstance(siginst2, SignalInstance)
    assert siginst1 is not siginst2

    # test that emitting from the deepcopied group doesn't affect the original
    siginst2.emit(1)
    mock.assert_not_called()
    mock2.assert_called_with(EmissionInfo(siginst2, (1,)))

    # test that emitting from the original group doesn't affect the deepcopied one
    mock2.reset_mock()
    siginst1.emit(1)
    mock.assert_called_with(EmissionInfo(siginst1, (1,)))
    mock2.assert_not_called()


def test_group_conflicts() -> None:
    with pytest.warns(UserWarning, match=r"Name \['connect'\] is reserved"):

        class MyGroup(SignalGroup):
            connect = Signal(int)  # type: ignore
            other_signal = Signal(int)

        class SubGroup(MyGroup):
            sig4 = Signal(int)

    assert "connect" in MyGroup._psygnal_signals
    assert "other_signal" in MyGroup._psygnal_signals
    group = MyGroup()
    assert isinstance(group["connect"], SignalInstance)
    assert not isinstance(group.connect, SignalInstance)

    with pytest.raises(
        TypeError,
        match="SignalGroup subclass cannot have attributes starting with '_psygnal'",
    ):

        class MyGroup2(SignalGroup):
            _psygnal_private = 1

    assert group.other_signal.name == "other_signal"
    assert group["connect"].name == "connect"

    subgroup = SubGroup()
    assert subgroup["connect"].name == "connect"
    assert subgroup.other_signal.name == "other_signal"


def test_group_iter() -> None:
    class Group1(SignalGroup):
        sig1 = Signal()
        sig2 = Signal()
        sig3 = Signal()

    # Delete Signal on Group class
    # You should never do that
    del Group1._psygnal_signals["sig1"]

    assert set(Group1._psygnal_signals) == {"sig2", "sig3"}
    assert hasattr(Group1, "sig1")

    g = Group1()

    # Delete Signal on Group instance
    # You should never do that
    del g._psygnal_signals["sig2"]

    assert "sig1" not in g
    assert "sig2" in g
    assert set(g) == {"sig2", "sig3"}

    with pytest.raises(KeyError):
        g["sig1"]

    sig1_t = g.sig1
    assert isinstance(sig1_t, SignalInstance)
    assert sig1_t.name == "sig1"

    sig2 = g["sig2"]
    assert isinstance(sig2, SignalInstance)
    assert sig2.name == "sig2"

    sig2_t = g.sig2
    assert isinstance(sig2_t, SignalInstance)
    assert sig2_t.name == "sig2"

    # Delete SignalInstance
    del g._psygnal_instances["sig3"]

    assert "sig3" not in g
    assert set(g) == {"sig2"}
    with pytest.raises(KeyError):
        g["sig3"]


def test_group_subclass() -> None:
    # Signals are passed to sub-classes
    class Group1(SignalGroup):
        sig1 = Signal()

    class Group2(Group1):
        sig2 = Signal()

    assert "sig1" in Group1._psygnal_signals
    assert "sig1" in Group2._psygnal_signals
    assert "sig2" in Group2._psygnal_signals
    assert "sig2" not in Group1._psygnal_signals

    assert hasattr(Group1, "sig1") and isinstance(Group1.sig1, Signal)
    assert hasattr(Group2, "sig1") and isinstance(Group2.sig1, Signal)
    assert hasattr(Group2, "sig2") and isinstance(Group2.sig2, Signal)
    assert not hasattr(Group1, "sig2")


def test_delayed_relay_connect() -> None:
    group = MyGroup()
    mock = Mock()
    gmock = Mock()
    assert len(group.sig1) == 0

    group.sig1.connect(mock)
    # group relay hasn't been connected to sig1 or sig2 yet
    assert len(group.sig1) == 1
    assert len(group.sig2) == 0

    group.all.connect(gmock)
    # NOW the relay is connected
    assert len(group.sig1) == 2
    assert len(group.sig2) == 1
    method = group.sig1._slots[-1].dereference()
    assert method
    assert method.__name__ == "_slot_relay"

    group.sig1.emit(1)
    mock.assert_called_once_with(1)
    gmock.assert_called_once_with(EmissionInfo(group.sig1, (1,)))

    group.all.disconnect(gmock)
    assert len(group.sig1) == 1
    assert len(group.all) == 0

    mock.reset_mock()
    gmock.reset_mock()
    group.sig1.emit(1)
    mock.assert_called_once_with(1)
    gmock.assert_not_called()


@pytest.mark.skipif(psygnal._compiled, reason="requires uncompiled psygnal")
def test_group_relay_signatures() -> None:
    from inspect import signature

    for name in dir(SignalGroup):
        if (
            hasattr(SignalRelay, name)
            and not name.startswith("_")
            and callable(getattr(SignalRelay, name))
        ):
            group_sig = signature(getattr(SignalGroup, name))
            relay_sig = signature(getattr(SignalRelay, name))

            assert group_sig == relay_sig


def test_group_relay_passthrough() -> None:
    group = MyGroup()

    mock1 = Mock()
    mock2 = Mock()

    # test connection
    group.connect(mock1)
    group.all.connect(mock2)
    group.sig1.emit(1)
    mock1.assert_called_once_with(EmissionInfo(group.sig1, (1,)))
    mock2.assert_called_once_with(EmissionInfo(group.sig1, (1,)))

    mock1.reset_mock()
    mock2.reset_mock()

    # test disconnection
    group.disconnect(mock1)
    group.all.disconnect(mock2)
    group.sig1.emit("hi")

    mock1.assert_not_called()
    mock2.assert_not_called()

    @group.connect(check_nargs=True)  # testing the decorator as well
    def _(x: int) -> None:
        mock1(x)

    group.all.connect(mock2)

    # test blocking
    with group.blocked():
        group.sig1.emit(1)

    mock1.assert_not_called()
    mock2.assert_not_called()

    with group.all.blocked():
        group.sig1.emit(1)

    mock1.assert_not_called()
    mock2.assert_not_called()

    # smoke test the rest
    group.connect_direct(mock1)
    group.block()
    group.unblock()
    group.blocked()
    group.pause()
    group.resume()
    group.paused()



import time
from collections.abc import Callable
from inspect import Parameter, signature
from unittest.mock import Mock

import pytest

from psygnal import SignalInstance, _compiled, debounced, throttled


def test_debounced() -> None:
    mock1 = Mock()
    f1 = debounced(mock1, timeout=10, leading=False)
    f2 = Mock()

    for _ in range(10):
        f1()
        f2()

    time.sleep(0.1)
    mock1.assert_called_once()

    assert f2.call_count == 10


def test_debounced_leading() -> None:
    mock1 = Mock()
    f1 = debounced(mock1, timeout=10, leading=True)
    f2 = Mock()

    for _ in range(10):
        f1()
        f2()

    time.sleep(0.1)
    assert mock1.call_count == 2
    assert f2.call_count == 10


def test_throttled() -> None:
    mock1 = Mock()
    f1 = throttled(mock1, timeout=10, leading=True)
    f2 = Mock()

    for _ in range(10):
        f1()
        f2()

    time.sleep(0.1)
    assert mock1.call_count == 2
    assert f2.call_count == 10


def test_throttled_trailing() -> None:
    mock1 = Mock()
    f1 = throttled(mock1, timeout=10, leading=False)
    f2 = Mock()

    for _ in range(10):
        f1()
        f2()

    time.sleep(0.1)
    assert mock1.call_count == 1
    assert f2.call_count == 10


def test_cancel() -> None:
    mock1 = Mock()
    f1 = debounced(mock1, timeout=50, leading=False)
    f1()
    f1()
    f1.cancel()
    time.sleep(0.2)
    mock1.assert_not_called()


def test_flush() -> None:
    mock1 = Mock()
    f1 = debounced(mock1, timeout=50, leading=False)
    f1()
    f1()
    f1.flush()
    time.sleep(0.2)
    mock1.assert_called_once()


@pytest.mark.parametrize("deco", [debounced, throttled])
def test_throttled_debounced_signature(deco: Callable) -> None:
    mock = Mock()

    @deco(timeout=0, leading=True)
    def f1(x: int) -> None:
        """Doc."""
        mock(x)

    # make sure we can still inspect the signature
    assert signature(f1).parameters["x"] == Parameter(
        "x", Parameter.POSITIONAL_OR_KEYWORD, annotation=int
    )

    # make sure these are connectable
    sig = SignalInstance((int, int, int))
    sig.connect(f1)
    sig.emit(1, 2, 3)
    mock.assert_called_once_with(1)

    if not _compiled:
        # unfortunately, dynamic assignment of __doc__ and stuff isn't possible in mypyc
        assert f1.__doc__ == "Doc."
        assert f1.__name__ == "f1"



from dataclasses import dataclass

import pytest
from attr import define

from psygnal import _dataclass_utils

try:
    from msgspec import Struct
except (ImportError, TypeError):  # type error on python 3.12-dev
    Struct = None  # type: ignore [assignment,misc]

try:
    from pydantic import __version__ as pydantic_version

    PYDANTIC2 = pydantic_version.startswith("2.")
except ImportError:
    PYDANTIC2 = False

VARIANTS = ["dataclass", "attrs_class", "pydantic_model"]
if Struct is not None:
    VARIANTS.append("msgspec_struct")


@pytest.mark.parametrize("frozen", [True, False], ids=["frozen", ""])
@pytest.mark.parametrize("type_", VARIANTS)
def test_dataclass_utils(type_: str, frozen: bool) -> None:
    if type_ == "attrs_class":

        @define(frozen=frozen)  # type: ignore
        class Foo:
            x: int
            y: str = "foo"

    elif type_ == "dataclass":

        @dataclass(frozen=frozen)  # type: ignore
        class Foo:  # type: ignore [no-redef]
            x: int
            y: str = "foo"

    elif type_ == "msgspec_struct":

        class Foo(Struct, frozen=frozen):  # type: ignore [no-redef]
            x: int
            y: str = "foo"

    elif type_ == "pydantic_model":
        pytest.importorskip("pydantic")
        from pydantic import BaseModel

        class Foo(BaseModel):  # type: ignore [no-redef]
            x: int
            y: str = "foo"

            if PYDANTIC2:
                model_config = {"frozen": frozen}

            else:

                class Config:
                    allow_mutation = not frozen

    for name in VARIANTS:
        is_type = getattr(_dataclass_utils, f"is_{name}")
        assert is_type(Foo) is (name == type_)
        assert is_type(Foo(x=1)) is (name == type_)

    assert list(_dataclass_utils.iter_fields(Foo)) == [("x", int), ("y", str)]

    if type_ == "msgspec_struct" and frozen:
        # not supported until next release of msgspec
        return

    assert _dataclass_utils.is_frozen(Foo) == frozen



import gc
import re
from functools import partial
from typing import Any
from unittest.mock import Mock
from weakref import ref

import pytest

from psygnal import SignalInstance
from psygnal._weak_callback import WeakCallback, weak_callback


@pytest.mark.parametrize(
    "type_",
    [
        "function",
        "toolz_function",
        "weak_func",
        "lambda",
        "method",
        "partial_method",
        "toolz_method",
        "setattr",
        "setitem",
        "mock",
        "weak_cb",
        "print",
    ],
)
def test_slot_types(type_: str, capsys: Any) -> None:
    mock = Mock()
    final_mock = Mock()

    class MyObj:
        def method(self, x: int) -> None:
            mock(x)
            return x

        def __setitem__(self, key, value):
            mock(value)
            return value

        def __setattr__(self, __name: str, __value) -> None:
            if __name == "x":
                mock(__value)
                return __value

    obj = MyObj()

    if type_ == "setattr":
        cb = weak_callback(setattr, obj, "x", finalize=final_mock)
    elif type_ == "setitem":
        cb = weak_callback(obj.__setitem__, "x", finalize=final_mock)
    elif type_ in {"function", "weak_func"}:

        def obj(x: int) -> None:
            mock(x)
            return x

        cb = weak_callback(obj, strong_func=(type_ == "function"), finalize=final_mock)
    elif type_ == "toolz_function":
        toolz = pytest.importorskip("toolz")

        @toolz.curry
        def obj(z: int, x: int) -> None:
            mock(x)
            return x

        cb = weak_callback(obj(5), finalize=final_mock)
    elif type_ == "lambda":
        cb = weak_callback(lambda x: mock(x) and x, finalize=final_mock)
    elif type_ == "method":
        cb = weak_callback(obj.method, finalize=final_mock)
    elif type_ == "partial_method":
        cb = weak_callback(partial(obj.method, 2), max_args=0, finalize=final_mock)
    elif type_ == "toolz_method":
        toolz = pytest.importorskip("toolz")
        cb = weak_callback(toolz.curry(obj.method, 2), max_args=0, finalize=final_mock)
    elif type_ == "mock":
        cb = weak_callback(mock, finalize=final_mock)
    elif type_ == "weak_cb":
        cb = weak_callback(obj.method, finalize=final_mock)
        cb = weak_callback(cb, finalize=final_mock)
    elif type_ == "print":
        cb = weak_callback(print, finalize=final_mock)

    assert isinstance(cb, WeakCallback)
    assert isinstance(cb.slot_repr(), str)
    cb.cb((2,))
    assert cb.dereference() is not None
    if type_ == "print":
        assert capsys.readouterr().out == "2\n"
        return

    mock.assert_called_once_with(2)
    mock.reset_mock()
    result = cb(2)
    if type_ not in ("setattr", "mock"):
        assert result == 2
    mock.assert_called_once_with(2)

    del obj

    if type_ not in ("function", "toolz_function", "lambda", "mock"):
        final_mock.assert_called_once_with(cb)
        assert cb.dereference() is None
        with pytest.raises(ReferenceError):
            cb.cb((2,))
        with pytest.raises(ReferenceError):
            cb(2)
    else:
        cb.cb((4,))
        mock.assert_called_with(4)


def test_weak_callable_equality() -> None:
    """Slot callers should be equal only if they represent the same bound-method."""

    class T:
        def x(self): ...

    t1 = T()
    t2 = T()
    t1_ref = ref(t1)
    t2_ref = ref(t2)

    bmt1_a = weak_callback(t1.x)
    bmt1_b = weak_callback(t1.x)
    bmt2_a = weak_callback(t2.x)
    bmt2_b = weak_callback(t2.x)

    assert bmt1_a != "not a weak callback"

    def _assert_equality() -> None:
        assert bmt1_a == bmt1_b
        assert bmt2_a == bmt2_b
        assert bmt1_a != bmt2_a
        assert bmt1_b != bmt2_b

    _assert_equality()
    del t1
    gc.collect()
    assert t1_ref() is None
    _assert_equality()
    del t2
    gc.collect()
    assert t2_ref() is None
    _assert_equality()


def test_nonreferencable() -> None:
    class T:
        __slots__ = ("x",)

        def method(self) -> None: ...

    t = T()
    with pytest.warns(UserWarning, match="failed to create weakref"):
        cb = weak_callback(t.method)
        assert cb.dereference() == t.method

    with pytest.raises(TypeError):
        weak_callback(t.method, on_ref_error="raise")

    cb = weak_callback(t.method, on_ref_error="ignore")
    assert cb.dereference() == t.method


@pytest.mark.parametrize("strong", [True, False])
def test_deref(strong: bool) -> None:
    def func(x): ...

    p = partial(func, 1)
    cb = weak_callback(p, strong_func=strong)
    dp = cb.dereference()

    assert dp.func is p.func
    assert dp.args == p.args
    assert dp.keywords == p.keywords


def test_queued_callbacks() -> None:
    from psygnal._queue import QueuedCallback

    def func(x):
        return x

    cb = weak_callback(func)
    qcb = QueuedCallback(cb, thread="current")

    assert qcb.dereference() is func
    assert qcb(1) == 1


def test_cb_raises() -> None:
    from psygnal import EmitLoopError

    sig = SignalInstance((int,), name="sig")

    class T:
        @property
        def x(self) -> int:
            return 1

        @x.setter
        def x(self, value: int) -> None:
            1 / value

        def __setitem__(self, key: str, value: int) -> Any:
            1 / value

        def method(self, x: int) -> None:
            1 / x

    t = T()

    sig.connect(t.method)
    error_re = re.compile(
        f"emitting signal.*'sig'.*{re.escape(__file__)}.*method", re.DOTALL
    )
    with pytest.raises(EmitLoopError, match=error_re):
        sig.emit("a")
    sig.disconnect(t.method)

    sig.connect_setattr(t, "x", maxargs=1)
    error_re = re.compile(
        f"emitting signal.*'sig'.*{re.escape(__file__)}.*x", re.DOTALL
    )
    with pytest.raises(EmitLoopError, match=error_re):
        sig.emit("a")
    sig.disconnect_setattr(t, "x")

    sig.connect_setitem(t, "x", maxargs=1)
    error_re = re.compile(
        f"emitting signal.*'sig'.*{re.escape(__file__)}.*__setitem__", re.DOTALL
    )
    with pytest.raises(EmitLoopError, match=error_re):
        sig.emit("a")
